/* encrypted p2p broadcast demo
 * pesco, 2009
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <assert.h>
#include "monty.h"
#include "byteorder.h"

#define LOCALPORT 1337
#define MAXPEERS 16
#define RECVBUFLEN 1024
#define SEENTBLLEN 1024
#define PKTIDLEN 8
#define HOSTIDLEN 8


/* IO functions */

void printbignum(FILE *f, int n, uint32_t *X)
{
        int i;

        for(i=0; i<n; i++) {
                fprintf(f, "%.8X", X[n-1-i]);
        }
	fprintf(f, "\n");
}

int readhexdigit(FILE *f)
{
	int c;

	for(;;) {
		c = getc(f);
		if(c>='0' && c<='9')
			return (c-'0');
		if(c>='a' && c<='f')
			return (c-'a'+10);
		if(c>='A' && c<='F')
			return (c-'A'+10);
	}
}
	
void readbignum(FILE *f, int n, uint32_t *X)
{
	int i,j;

	for(i=n-1; i>=0; i--) {
		uint32_t x=0;

		for(j=0; j<8; j++) {
			x <<= 4;
			x |= readhexdigit(f);
		}

		X[i] = x;
	}
}

int main(int argc, char **argv)
{
        /* 1536-bit modulus from RFC 3526, little-endian */
        uint32_t N[] = {0xFFFFFFFF, 0xFFFFFFFF, 0xCA237327, 0xF1746C08, 0x4ABC9804, 0x670C354E,
                        0x7096966D, 0x9ED52907, 0x208552BB, 0x1C62F356, 0xDCA3AD96, 0x83655D23,
                        0xFD24CF5F, 0x69163FA8, 0x1C55D39A, 0x98DA4836, 0xA163BF05, 0xC2007CB8,
                        0xECE45B3D, 0x49286651, 0x7C4B1FE6, 0xAE9F2411, 0x5A899FA5, 0xEE386BFB,
                        0xF406B7ED, 0x0BFF5CB6, 0xA637ED6B, 0xF44C42E9, 0x625E7EC6, 0xE485B576,
                        0x6D51C245, 0x4FE1356D, 0xF25F1437, 0x302B0A6D, 0xCD3A431B, 0xEF9519B3,
                        0x8E3404DD, 0x514A0879, 0x3B139B22, 0x020BBEA6, 0x8A67CC74, 0x29024E08,
                        0x80DC1CD1, 0xC4C6628B, 0x2168C234, 0xC90FDAA2, 0xFFFFFFFF, 0xFFFFFFFF};
        /* 2^3072 mod N */
        uint32_t RR[] = {0x32c695e0, 0xf115d27d, 0x67478c73, 0x8e0e3e21, 0x8397f245, 0xd0ab92e1,
                         0xbcd49d68, 0xf466ee5f, 0x3b01e018, 0x8f2331b1, 0x98b5fb62, 0x7e8cd2ac,
                         0x7a58f170, 0xb9052bb4, 0xdb102d39, 0xb004a750, 0x93ae1ceb, 0x04a541ff,
                         0x8e434130, 0x07cd0a62, 0x04b9f796, 0x1c729c7e, 0x196b7e88, 0xb8fe6121,
                         0x0223b76b, 0x8e1abd78, 0xd46fec23, 0x22c296e9, 0xb270521b, 0xd62a0eea,
                         0xd4053f54, 0xdc541a4e, 0x969b7f02, 0xf8056564, 0xa87c7b37, 0x0be49647,
                         0x67984460, 0x57b59348, 0x9a36a51f, 0x102630fa, 0xcc2456ef, 0xe9c3fa02,
                         0x7929a1c7, 0xae594104, 0x6cc1ebd2, 0xee9c9a21, 0x59541c01, 0xe3b33c72};
        int n = sizeof(N) / sizeof(uint32_t);
	uint32_t R[n];
        uint32_t one[n];

        uint32_t g[n];          /* generator = 2 */
        uint32_t gR[n];         /* generator, montgomery representation */
	uint32_t a[n];          /* random number, alice */
	uint32_t b[n];          /* random number, bob */
	uint32_t gaR[n];        /* g^a, montgomery rep */
	uint32_t gbR[n];        /* g^b, montgomery rep */
	uint32_t gabR[n];       /* g^ab, montgomery rep */
	uint32_t mR[n];         /* message, montgomery rep */
	uint32_t m[n];          /* message, converted back */

	FILE *fpriv, *fpub, *fpeers;
	int npeers, i, j;
	struct sockaddr_in me;
	struct sockaddr_in peers[MAXPEERS];
	int sockfd;

	unsigned char pkt[RECVBUFLEN];
	struct sockaddr_in from;
	socklen_t fromlen;
	ssize_t pktlen;
	unsigned char seen[SEENTBLLEN][PKTIDLEN];


	/* initialize */
        memset(one,  0, sizeof(N)); one[0] = 1;
        memset(g,    0, sizeof(N)); g[0] = 2;

	/* convert stuff to montgomery representation */
	monty_mul(n, N, R, one, RR);
        monty_mul(n, N, gR, g, RR);


	/* acquire key material */
	if(!(fpriv = fopen("privkey", "r")))
		perror("privkey");
	if(!(fpub = fopen("pubkey", "r")))
		perror("pubkey");
	if(!fpriv && !fpub) {
		/* no key there => generate one */
		fpriv = fopen("privkey", "w");
		if(fpriv == NULL) {
			perror("privkey");
			return -1;
		}
		fpub = fopen("pubkey", "w");
		if(fpub == NULL) {
			perror("pubkey");
			return -1;
		}
		printf("no host key found, generating...\n");
		mrand(n, N, a);
		monty_exp(n, N, R, gaR, gR, a);  /* compute g^a */
		printbignum(fpriv, n, a);
		printbignum(fpub, n, gaR);
	} else if(!fpriv) {
		/* missing private key => huh? */
		printf("private key file missing - bailing out\n");
		printf("recover file 'privkey' or remove 'pubkey' to generate a new key\n");
		return 1;
	} else if(!fpub) {
		/* missing public key => recover from private key */
		printf("public key file missing - recalculating\n");
		readbignum(fpriv, n, a);
		monty_exp(n, N, R, gaR, gR, a);  /* compute g^a */
		fpub = fopen("pubkey", "w");
		if(fpub == NULL) {
			perror("pubkey");
			return -1;
		}
		printbignum(fpub, n, gaR);
	} else {
		readbignum(fpriv, n, a);
		readbignum(fpub, n, gaR);
		printf("host key loaded\n");
	}
	fclose(fpriv);
	fclose(fpub);

	/* read peers file */
	fpeers = fopen("peers", "r");
	if(fpeers == NULL) {
		perror("peers");
		return -1;
	}
	npeers = 0;
	for(i=1; !feof(fpeers) && npeers<MAXPEERS; i++) {
		char line[256];
		char ip[16];
		int port;
		struct sockaddr_in *peer = peers+npeers;

		if(!fgets(line, 256, fpeers))
			break;
		if(sscanf(line, "%15s %d", ip, &port) < 2) {
			printf("peers:%d: parse error\n", i);
			continue;
		}

		peer->sin_family = AF_INET;
		if(!inet_aton(ip, &peer->sin_addr)) {
			printf("peers:%d: invalid address '%s'\n", i, ip);
			continue;
		}
		peer->sin_port = htons(port);
		npeers++;
	}
	fclose(fpeers);
	printf("%d peers loaded\n", npeers);

	/* open receiving socket */
	sockfd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if(sockfd < 0) {
		perror("socket");
		return -1;
	}
	me.sin_family = AF_INET;
	me.sin_addr.s_addr = 0;  /* a.k.a. INADDR_ANY */
	me.sin_port = htons(LOCALPORT);
	if(bind(sockfd, (struct sockaddr *)&me, sizeof(me)) < 0) {
		perror("bind");
		return -1;
	}
	printf("receiving packets on port %d\n", LOCALPORT);
	printf("CAUTION: Message senders can be spoofed.\n");

	assert(PKTIDLEN < n);
	assert(HOSTIDLEN < n);
	memset(seen, 0, sizeof(seen));
	i=0;
	while(1) {
		/* wait for packet */
		fromlen = sizeof(from);
		pktlen = recvfrom(sockfd, pkt, RECVBUFLEN, 0,
			(struct sockaddr*)&from, &fromlen);
		if(pktlen<0) {
			perror("recvfrom");
			continue;
		}
		assert(fromlen == sizeof(from));

		/* packet sanity checks */
		if(pktlen<2*n*4) {
			printf("received packet shorter than expected, ignored\n");
			continue;
		}
		if(pktlen>2*n*4) {
			printf("received packet longer than expected, ignored\n");
			continue;
		}
		
		/* discard packet if already seen */
		for(j=0; j<SEENTBLLEN; j++) {
			if(!memcmp(pkt, seen[j], PKTIDLEN))
				break;
		}
		if(j<SEENTBLLEN) {
			printf("packet already seen (%d), ignored\n", j);
			continue;
		}
		memcpy(seen[i], pkt, PKTIDLEN);

		/* input marshaling */
		memcpy(gbR, pkt, n*4);
		memcpy(mR, pkt+n*4, n*4);
		ltoh(gbR, n*2);

		/* decrypt */
		monty_exp(n, N, R, gabR, gbR, a);
		sub(n, mR, 0, gabR);
		monty_mul(n, N, m, mR, one);

		/* if first few bytes match our public key,
		 * this packet is for us. */
		if(!memcmp(m, gaR, HOSTIDLEN)) {
			/* on match: extract and display message */
			char *msg = (char *)m + HOSTIDLEN;
			htol(m,n);

			msg[n*4 - HOSTIDLEN - 1] = '\0';
			printf("%s\n", msg);
		} else {
			/* no match: forward packet to all peers */
			for(j=0; j<npeers; j++) {
				ssize_t r;
				r = sendto(sockfd, pkt, pktlen, 0,
					(struct sockaddr*)(peers+j),
					sizeof(struct sockaddr_in));
				if(r < 0)
					perror("sendto");
			}
		}

		i = (i+1) % SEENTBLLEN;
	}

	close(sockfd);
	return 0;
}
